import { __values, __assign, __awaiter, __generator, __extends } from 'tslib';
import { inject } from '@angular/core';
import { AngularFirestore } from '@angular/fire/firestore';
import { withTransaction, updateEntities, arrayAdd, arrayUpsert, arrayRemove, arrayUpdate } from '@datorama/akita';
import { isObservable, of, combineLatest, race } from 'rxjs';
import { map, switchMap, tap, finalize, distinctUntilChanged, filter, shareReplay } from 'rxjs/operators';
import { Router } from '@angular/router';
import { RouterQuery } from '@datorama/akita-ng-router-store';

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
// Helper to retrieve the id and path of a document in the collection
/**
 * @param {?} options
 * @param {?=} collectionPath
 * @return {?}
 */
function getIdAndPath(options, collectionPath) {
    /** @type {?} */
    var path = '';
    /** @type {?} */
    var id = '';
    if (options['id']) {
        if (!collectionPath) {
            throw new Error('You should provide the colletion path with the id');
        }
        id = options['id'];
        path = collectionPath + "/" + id;
    }
    else if (options['path']) {
        path = options['path'];
        /** @type {?} */
        var part = path.split('/');
        if (path.length % 2 !== 0) {
            throw new Error("Path " + path + " doesn't look like a Firestore's document path");
        }
        id = part[part.length - 1];
    }
    else {
        throw new Error("You should provide either an \"id\" OR a \"path\".");
    }
    return { id: id, path: path };
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
// Credit to @arielgueta https://dev.to/arielgueta/getting-started-with-akita-and-firebase-3pe2
/**
 * @template S
 * @this {?}
 * @param {?} actions
 * @return {?}
 */
function syncFromAction(actions) {
    var e_1, _a, _b;
    this.store.setLoading(false);
    if (actions.length === 0) {
        return;
    }
    try {
        for (var actions_1 = __values(actions), actions_1_1 = actions_1.next(); !actions_1_1.done; actions_1_1 = actions_1.next()) {
            var action = actions_1_1.value;
            /** @type {?} */
            var id = (/** @type {?} */ (action.payload.doc.id));
            /** @type {?} */
            var entity = action.payload.doc.data();
            switch (action.type) {
                case 'added': {
                    this.store.upsert(id, __assign((_b = {}, _b[this.idKey] = id, _b), entity));
                    break;
                }
                case 'removed': {
                    this.store.remove(id);
                    break;
                }
                case 'modified': {
                    this.store.update(id, entity);
                }
            }
        }
    }
    catch (e_1_1) { e_1 = { error: e_1_1 }; }
    finally {
        try {
            if (actions_1_1 && !actions_1_1.done && (_a = actions_1.return)) _a.call(actions_1);
        }
        finally { if (e_1) throw e_1.error; }
    }
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/**
 * @template S
 */
var  /**
 * @template S
 */
CollectionService = /** @class */ (function () {
    function CollectionService(store, pathToCollection) {
        this.store = store;
        this.pathToCollection = pathToCollection;
        if (!this.constructor['path'] && !this.pathToCollection) {
            throw new Error('You should provide a path to the collection');
        }
        try {
            this.db = inject(AngularFirestore);
        }
        catch (err) {
            throw new Error('CollectionService requires AngularFirestore.');
        }
    }
    Object.defineProperty(CollectionService.prototype, "idKey", {
        get: /**
         * @return {?}
         */
        function () {
            return this.constructor['idKey'] || this.store.idKey;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(CollectionService.prototype, "path", {
        /** The path to the collection in Firestore */
        get: /**
         * The path to the collection in Firestore
         * @return {?}
         */
        function () {
            return this.constructor['path'] || this.pathToCollection;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(CollectionService.prototype, "currentPath", {
        /** A snapshot of the path */
        get: /**
         * A snapshot of the path
         * @return {?}
         */
        function () {
            if (isObservable(this.path)) {
                throw new Error('Cannot get a snapshot of the path if it is an Observable');
            }
            return this.path;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(CollectionService.prototype, "path$", {
        /** An observable version of the path */
        get: /**
         * An observable version of the path
         * @return {?}
         */
        function () {
            return isObservable(this.path) ? this.path : of(this.path);
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(CollectionService.prototype, "collection", {
        /**
         * The Angular Fire collection
         * @notice If path is an observable, it becomes an observable.
         */
        get: /**
         * The Angular Fire collection
         * \@notice If path is an observable, it becomes an observable.
         * @return {?}
         */
        function () {
            return this.db.collection(this.currentPath);
        },
        enumerable: true,
        configurable: true
    });
    /** Preformat the document before updating Firestore */
    /**
     * Preformat the document before updating Firestore
     * @protected
     * @template E
     * @param {?} document
     * @return {?}
     */
    CollectionService.prototype.preFormat = /**
     * Preformat the document before updating Firestore
     * @protected
     * @template E
     * @param {?} document
     * @return {?}
     */
    function (document) {
        return document;
    };
    Object.defineProperty(CollectionService.prototype, "config", {
        /** The config given by the `CollectonConfig` */
        get: /**
         * The config given by the `CollectonConfig`
         * @return {?}
         */
        function () {
            return {
                path: this.constructor['path'],
                idKey: this.constructor['idKey']
            };
        },
        enumerable: true,
        configurable: true
    });
    /**
     * @param {?=} pathOrQuery
     * @param {?=} queryFn
     * @return {?}
     */
    CollectionService.prototype.syncCollection = /**
     * @param {?=} pathOrQuery
     * @param {?=} queryFn
     * @return {?}
     */
    function (pathOrQuery, queryFn) {
        var _this = this;
        if (pathOrQuery === void 0) { pathOrQuery = this.path; }
        /** @type {?} */
        var path;
        if (isObservable(pathOrQuery)) {
            path = pathOrQuery;
        }
        else if (typeof pathOrQuery === 'function') {
            queryFn = pathOrQuery;
            path = this.path$;
        }
        else {
            path = this.path$;
        }
        // If there is no entity yet set loading
        if (!this.store._value().ids || this.store._value().ids.length === 0) {
            this.store.setLoading(true);
        }
        // Start Listening
        return path.pipe(map((/**
         * @param {?} collectionPath
         * @return {?}
         */
        function (collectionPath) { return _this.db.collection(collectionPath, queryFn); })), switchMap((/**
         * @param {?} collection
         * @return {?}
         */
        function (collection) { return collection.stateChanges(); })), withTransaction(syncFromAction.bind(this)));
    };
    /**
     * @param {?=} idOrQuery
     * @param {?=} queryGroupFn
     * @return {?}
     */
    CollectionService.prototype.syncCollectionGroup = /**
     * @param {?=} idOrQuery
     * @param {?=} queryGroupFn
     * @return {?}
     */
    function (idOrQuery, queryGroupFn) {
        /** @type {?} */
        var path = (typeof idOrQuery === 'string') ? idOrQuery : this.currentPath;
        /** @type {?} */
        var query = typeof idOrQuery === 'function' ? idOrQuery : queryGroupFn;
        /** @type {?} */
        var collectionId = path.split('/').pop();
        return this.db.collectionGroup(collectionId, query)
            .stateChanges()
            .pipe(withTransaction(syncFromAction.bind(this)));
    };
    /**
     * Sync the store with several documents
     * @param ids An array of ids
     */
    /**
     * Sync the store with several documents
     * @param {?} ids An array of ids
     * @return {?}
     */
    CollectionService.prototype.syncManyDocs = /**
     * Sync the store with several documents
     * @param {?} ids An array of ids
     * @return {?}
     */
    function (ids) {
        var _this = this;
        /** @type {?} */
        var syncs = ids.map((/**
         * @param {?} id
         * @return {?}
         */
        function (id) { return _this.path$.pipe(map((/**
         * @param {?} collectionPath
         * @return {?}
         */
        function (collectionPath) { return getIdAndPath({ id: id }, collectionPath); })), switchMap((/**
         * @param {?} __0
         * @return {?}
         */
        function (_a) {
            var path = _a.path;
            return _this.db.doc(path).valueChanges();
        })), map((/**
         * @param {?} doc
         * @return {?}
         */
        function (doc) {
            var _a;
            return ((/** @type {?} */ (__assign((_a = {}, _a[_this.idKey] = id, _a), doc))));
        }))); }));
        return combineLatest(syncs).pipe(tap((/**
         * @param {?} entities
         * @return {?}
         */
        function (entities) { return _this.store.upsertMany(entities); })));
    };
    /**
     * Stay in sync with one document
     * @param options An object with EITHER `id` OR `path`.
     * @note We need to use id and path because there is no way to differentiate them.
     */
    /**
     * Stay in sync with one document
     * \@note We need to use id and path because there is no way to differentiate them.
     * @param {?} options An object with EITHER `id` OR `path`.
     * @return {?}
     */
    CollectionService.prototype.syncDoc = /**
     * Stay in sync with one document
     * \@note We need to use id and path because there is no way to differentiate them.
     * @param {?} options An object with EITHER `id` OR `path`.
     * @return {?}
     */
    function (options) {
        var _this = this;
        /** @type {?} */
        var id;
        return this.path$.pipe(map((/**
         * @param {?} collectionPath
         * @return {?}
         */
        function (collectionPath) { return getIdAndPath(options, collectionPath); })), tap((/**
         * @param {?} data
         * @return {?}
         */
        function (data) {
            id = (/** @type {?} */ (data.id));
            if (!_this.store._value().ids.includes(id)) {
                _this.store.setLoading(true);
            }
        })), switchMap((/**
         * @param {?} __0
         * @return {?}
         */
        function (_a) {
            var path = _a.path;
            return _this.db.doc(path).valueChanges();
        })), map((/**
         * @param {?} entity
         * @return {?}
         */
        function (entity) {
            var _a;
            /** @type {?} */
            var data = __assign((_a = {}, _a[_this.idKey] = id, _a), entity);
            _this.store.upsert(id, data);
            _this.store.setLoading(false);
            return data;
        })));
    };
    /**
     * @param {?} options
     * @return {?}
     */
    CollectionService.prototype.syncActive = /**
     * @param {?} options
     * @return {?}
     */
    function (options) {
        var _this = this;
        if (Array.isArray(options)) {
            return this.syncManyDocs(options).pipe(tap((/**
             * @param {?} _
             * @return {?}
             */
            function (_) { return _this.store.setActive((/** @type {?} */ (options))); })));
        }
        else {
            return this.syncDoc(options).pipe(tap((/**
             * @param {?} entity
             * @return {?}
             */
            function (entity) { return _this.store.setActive(entity[_this.idKey]); })));
        }
    };
    /**
     * @param {?=} idOrQuery
     * @return {?}
     */
    CollectionService.prototype.getValue = /**
     * @param {?=} idOrQuery
     * @return {?}
     */
    function (idOrQuery) {
        return __awaiter(this, void 0, void 0, function () {
            var snapshot, snapshot;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        if (!(typeof idOrQuery === 'string')) return [3 /*break*/, 2];
                        return [4 /*yield*/, this.db.doc(this.currentPath + "/" + idOrQuery).ref.get()];
                    case 1:
                        snapshot = _a.sent();
                        return [2 /*return*/, (/** @type {?} */ (snapshot.data()))];
                    case 2: return [4 /*yield*/, this.db.collection(this.currentPath, idOrQuery).ref.get()];
                    case 3:
                        snapshot = _a.sent();
                        return [2 /*return*/, snapshot.docs.map((/**
                             * @param {?} doc
                             * @return {?}
                             */
                            function (doc) { return (/** @type {?} */ (doc.data())); }))];
                }
            });
        });
    };
    /**
     * Add a document or a list of document to Firestore
     * @param docs A document or a list of document
     * @param write batch or transaction to run the operation into
     */
    /**
     * Add a document or a list of document to Firestore
     * @param {?} documents
     * @param {?=} write batch or transaction to run the operation into
     * @return {?}
     */
    CollectionService.prototype.add = /**
     * Add a document or a list of document to Firestore
     * @param {?} documents
     * @param {?=} write batch or transaction to run the operation into
     * @return {?}
     */
    function (documents, write) {
        return __awaiter(this, void 0, void 0, function () {
            var addId, addDocuments, batch_1;
            var _this = this;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        // Add id to the document
                        addId = (/**
                         * @param {?} doc
                         * @return {?}
                         */
                        function (doc) {
                            var _a;
                            return (__assign({}, doc, (_a = {}, _a[_this.idKey] = doc[_this.idKey] || _this.db.createId(), _a)));
                        });
                        // Add the list of document with the right atomic operation
                        addDocuments = (/**
                         * @param {?} operation
                         * @param {?} atomicWrite
                         * @return {?}
                         */
                        function (operation, atomicWrite) {
                            /** @type {?} */
                            var docs = Array.isArray(documents) ? documents : [documents];
                            /** @type {?} */
                            var operations = docs.map((/**
                             * @param {?} document
                             * @return {?}
                             */
                            function (document) {
                                /** @type {?} */
                                var doc = addId(_this.preFormat(document));
                                var ref = _this.db.doc(_this.currentPath + "/" + doc[_this.idKey]).ref;
                                if (_this.onCreate) {
                                    _this.onCreate(doc, atomicWrite);
                                }
                                return operation(ref, doc);
                            }));
                            return Promise.all(operations);
                        });
                        if (!write) return [3 /*break*/, 1];
                        return [2 /*return*/, addDocuments((/**
                             * @param {?} ref
                             * @param {?} doc
                             * @return {?}
                             */
                            function (ref, doc) { return write.set(ref, doc); }), write)];
                    case 1:
                        batch_1 = this.db.firestore.batch();
                        return [4 /*yield*/, addDocuments((/**
                             * @param {?} ref
                             * @param {?} doc
                             * @return {?}
                             */
                            function (ref, doc) { return batch_1.set(ref, doc); }), batch_1)];
                    case 2:
                        _a.sent();
                        return [2 /*return*/, batch_1.commit()];
                }
            });
        });
    };
    /**
     * Remove one or several document from Firestore
     * @param id A unique or list of id representing the document
     * @param write batch or transaction to run the operation into
     */
    /**
     * Remove one or several document from Firestore
     * @param {?} id A unique or list of id representing the document
     * @param {?=} write batch or transaction to run the operation into
     * @return {?}
     */
    CollectionService.prototype.remove = /**
     * Remove one or several document from Firestore
     * @param {?} id A unique or list of id representing the document
     * @param {?=} write batch or transaction to run the operation into
     * @return {?}
     */
    function (id, write) {
        return __awaiter(this, void 0, void 0, function () {
            var removeDocuments, batch_2;
            var _this = this;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        removeDocuments = (/**
                         * @param {?} operation
                         * @param {?} atomicWrite
                         * @return {?}
                         */
                        function (operation, atomicWrite) {
                            /** @type {?} */
                            var ids = Array.isArray(id) ? id : [id];
                            /** @type {?} */
                            var operations = ids.map((/**
                             * @param {?} docId
                             * @return {?}
                             */
                            function (docId) { return __awaiter(_this, void 0, void 0, function () {
                                var ref;
                                return __generator(this, function (_a) {
                                    switch (_a.label) {
                                        case 0:
                                            ref = this.db.doc(this.currentPath + "/" + docId).ref;
                                            if (!this.onDelete) return [3 /*break*/, 2];
                                            return [4 /*yield*/, this.onDelete(docId, atomicWrite)];
                                        case 1:
                                            _a.sent();
                                            _a.label = 2;
                                        case 2: return [2 /*return*/, operation(ref)];
                                    }
                                });
                            }); }));
                            return Promise.all(operations);
                        });
                        if (!write) return [3 /*break*/, 1];
                        return [2 /*return*/, removeDocuments((/**
                             * @param {?} ref
                             * @return {?}
                             */
                            function (ref) { return write.delete(ref); }), write)];
                    case 1:
                        batch_2 = this.db.firestore.batch();
                        return [4 /*yield*/, removeDocuments((/**
                             * @param {?} ref
                             * @return {?}
                             */
                            function (ref) { return batch_2.delete(ref); }), batch_2)];
                    case 2:
                        _a.sent();
                        return [2 /*return*/, batch_2.commit()];
                }
            });
        });
    };
    /**
     * @param {?} idsOrFn
     * @param {?=} stateFnOrWrite
     * @param {?=} write
     * @return {?}
     */
    CollectionService.prototype.update = /**
     * @param {?} idsOrFn
     * @param {?=} stateFnOrWrite
     * @param {?=} write
     * @return {?}
     */
    function (idsOrFn, stateFnOrWrite, write) {
        return __awaiter(this, void 0, void 0, function () {
            var isEntity, ids, newStateOrFn, state_1, updates, updateDocuments, batch_3;
            var _this = this;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        isEntity = (/**
                         * @param {?} value
                         * @return {?}
                         */
                        function (value) {
                            return typeof value === 'object' && idsOrFn[_this.idKey];
                        });
                        // Set variables
                        ids = [];
                        newStateOrFn = (/** @type {?} */ (stateFnOrWrite));
                        if (isEntity(idsOrFn)) {
                            ids = [idsOrFn[this.idKey]];
                            write = (/** @type {?} */ (stateFnOrWrite));
                            newStateOrFn = idsOrFn;
                        }
                        else if (typeof idsOrFn === 'function') {
                            state_1 = this.store._value();
                            ids = (/** @type {?} */ (state_1.ids.filter((/**
                             * @param {?} id
                             * @return {?}
                             */
                            function (id) { return idsOrFn(state_1.entities[id]); }))));
                        }
                        else {
                            ids = Array.isArray(idsOrFn) ? idsOrFn : [idsOrFn];
                        }
                        if (!Array.isArray(ids) || ids.length === 0) {
                            return [2 /*return*/];
                        }
                        // Create updates
                        updates = updateEntities({
                            state: this.store._value(),
                            ids: ids,
                            idKey: this.idKey,
                            newStateOrFn: newStateOrFn,
                            preUpdateEntity: this.store.akitaPreUpdateEntity
                        }).entities;
                        // Update the list of document with the right atomic operation
                        updateDocuments = (/**
                         * @param {?} operation
                         * @param {?} atomicWrite
                         * @return {?}
                         */
                        function (operation, atomicWrite) {
                            /** @type {?} */
                            var operations = Object.keys(updates).map((/**
                             * @param {?} key
                             * @return {?}
                             */
                            function (key) { return __awaiter(_this, void 0, void 0, function () {
                                var doc, ref;
                                return __generator(this, function (_a) {
                                    switch (_a.label) {
                                        case 0:
                                            doc = this.preFormat(updates[key]);
                                            ref = this.db.doc(this.currentPath + "/" + doc[this.idKey]).ref;
                                            if (!this.onUpdate) return [3 /*break*/, 2];
                                            return [4 /*yield*/, this.onUpdate(doc, atomicWrite)];
                                        case 1:
                                            _a.sent();
                                            _a.label = 2;
                                        case 2: return [2 /*return*/, operation(ref, doc)];
                                    }
                                });
                            }); }));
                            return Promise.all(operations);
                        });
                        if (!write) return [3 /*break*/, 1];
                        return [2 /*return*/, updateDocuments((/**
                             * @param {?} ref
                             * @param {?} doc
                             * @return {?}
                             */
                            function (ref, doc) { return write.update(ref, doc); }), write)];
                    case 1:
                        batch_3 = this.db.firestore.batch();
                        return [4 /*yield*/, updateDocuments((/**
                             * @param {?} ref
                             * @param {?} doc
                             * @return {?}
                             */
                            function (ref, doc) { return batch_3.update(ref, doc); }), batch_3)];
                    case 2:
                        _a.sent();
                        return [2 /*return*/, batch_3.commit()];
                }
            });
        });
    };
    return CollectionService;
}());

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/**
 * Set the configuration for the collection service
 * @param {?=} options
 * @return {?}
 */
function CollectionConfig(options) {
    if (options === void 0) { options = {}; }
    return (/**
     * @param {?} constructor
     * @return {?}
     */
    function (constructor) {
        Object.keys(options).forEach((/**
         * @param {?} key
         * @return {?}
         */
        function (key) { return constructor[key] = options[key]; }));
    });
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/**
 * @param {?} data
 * @return {?}
 */
function CollectionGuardConfig(data) {
    return (/**
     * @param {?} constructor
     * @return {?}
     */
    function (constructor) {
        Object.keys(data).forEach((/**
         * @param {?} key
         * @return {?}
         */
        function (key) { return (constructor[key] = data[key]); }));
    });
}
/**
 * @template S
 */
var  /**
 * @template S
 */
CollectionGuard = /** @class */ (function () {
    function CollectionGuard(service) {
        this.service = service;
        try {
            this.router = inject(Router);
        }
        catch (err) {
            throw new Error('CollectionGuard requires RouterModule to be imported');
        }
    }
    Object.defineProperty(CollectionGuard.prototype, "awaitSync", {
        // Can be override by the extended class
        /** Should the guard wait for connection to Firestore to complete */
        get: 
        // Can be override by the extended class
        /**
         * Should the guard wait for connection to Firestore to complete
         * @protected
         * @return {?}
         */
        function () {
            return this.constructor['awaitSync'] || false;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(CollectionGuard.prototype, "queryFn", {
        // Can be override by the extended class
        /** A queryFn to execute for sync */
        get: 
        // Can be override by the extended class
        /**
         * A queryFn to execute for sync
         * @protected
         * @return {?}
         */
        function () {
            return this.constructor['queryFn'];
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(CollectionGuard.prototype, "redirect", {
        // Can be override by the extended class
        /** The route to redirect to if you sync failed */
        get: 
        // Can be override by the extended class
        /**
         * The route to redirect to if you sync failed
         * @protected
         * @return {?}
         */
        function () {
            return this.constructor['redirect'];
        },
        enumerable: true,
        configurable: true
    });
    // Can be override by the extended class
    /** The method to subscribe to while route is active */
    // Can be override by the extended class
    /**
     * The method to subscribe to while route is active
     * @protected
     * @param {?} next
     * @return {?}
     */
    CollectionGuard.prototype.sync = 
    // Can be override by the extended class
    /**
     * The method to subscribe to while route is active
     * @protected
     * @param {?} next
     * @return {?}
     */
    function (next) {
        var _a = (/** @type {?} */ (next.data)).queryFn, queryFn = _a === void 0 ? this.queryFn : _a;
        return this.service.syncCollection(queryFn);
    };
    /**
     * @param {?} next
     * @return {?}
     */
    CollectionGuard.prototype.canActivate = /**
     * @param {?} next
     * @return {?}
     */
    function (next) {
        var _this = this;
        var _a = (/** @type {?} */ (next.data)), _b = _a.redirect, redirect = _b === void 0 ? this.redirect : _b, _c = _a.awaitSync, awaitSync = _c === void 0 ? this.awaitSync : _c;
        return new Promise((/**
         * @param {?} res
         * @param {?} rej
         * @return {?}
         */
        function (res, rej) {
            if (awaitSync) {
                _this.subscription = _this.sync(next).subscribe({
                    next: (/**
                     * @param {?} result
                     * @return {?}
                     */
                    function (result) {
                        switch (typeof result) {
                            case 'string': {
                                _this.subscription.unsubscribe();
                                return res(_this.router.parseUrl(result));
                            }
                            case 'boolean':
                                return res(result);
                            default:
                                return res(true);
                        }
                    }),
                    error: (/**
                     * @param {?} err
                     * @return {?}
                     */
                    function (err) {
                        _this.subscription.unsubscribe();
                        res(_this.router.parseUrl(redirect || ''));
                        throw new Error(err);
                    })
                });
            }
            else {
                _this.subscription = _this.sync(next).subscribe();
                res(true);
            }
        }));
    };
    /**
     * @return {?}
     */
    CollectionGuard.prototype.canDeactivate = /**
     * @return {?}
     */
    function () {
        if (this.subscription) {
            this.subscription.unsubscribe();
        }
        return true;
    };
    return CollectionGuard;
}());

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/**
 * @param {?} __0
 * @return {?}
 */
function shoudCancel(_a) {
    var validate = _a.validate, cancel = _a.cancel;
    return race([validate.pipe(map((/**
         * @param {?} _
         * @return {?}
         */
        function (_) { return false; }))), cancel.pipe(map((/**
         * @param {?} _
         * @return {?}
         */
        function (_) { return true; })))]);
}
/**
 * @param {?} __0
 * @return {?}
 */
function waitForCancel(_a) {
    var startWith = _a.startWith, endWith = _a.endWith, shouldValidate = _a.shouldValidate, shouldCancel = _a.shouldCancel;
    return __awaiter(this, void 0, void 0, function () {
        var cancelled;
        return __generator(this, function (_b) {
            switch (_b.label) {
                case 0:
                    startWith();
                    return [4 /*yield*/, race([
                            shouldValidate.pipe(map((/**
                             * @param {?} _
                             * @return {?}
                             */
                            function (_) { return false; }))),
                            shouldCancel.pipe(map((/**
                             * @param {?} _
                             * @return {?}
                             */
                            function (_) { return true; })))
                        ]).toPromise()];
                case 1:
                    cancelled = _b.sent();
                    endWith(cancelled);
                    return [2 /*return*/];
            }
        });
    });
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/**
 * @template E, K
 * @param {?} query
 * @param {?} parent
 * @return {?}
 */
function getSubQuery(query, parent) {
    if (typeof query !== 'function') {
        return query;
    }
    return query(parent);
}
/**
 * @param {?} path
 * @return {?}
 */
function isDocPath(path) {
    return path.split('/').length % 2 === 0;
}
/**
 * Transform a path into a collection Query
 * @template T
 * @param {?} path
 * @param {?=} queryFn
 * @return {?}
 */
function collection(path, queryFn) {
    return (/** @type {?} */ ({ path: path, queryFn: queryFn }));
}
/**
 * Transform a path into a doc query
 * @template T
 * @param {?} path
 * @return {?}
 */
function doc(path) {
    return (/** @type {?} */ ({ path: path }));
}
/**
 * Check if a value is a query
 * @template T
 * @param {?} query
 * @return {?}
 */
function isQuery(query) {
    if (typeof query === 'object') {
        return !!query['path'];
    }
    return false;
}
/**
 * Sync the collection
 * @template E
 * @this {?}
 * @param {?} query The query to trigger
 * @return {?}
 */
function syncQuery(query) {
    var _this = this;
    // If single query
    if (typeof query === 'string') {
        return isDocPath(query)
            ? this.syncDoc({ path: query })
            : this.syncCollection(query);
    }
    if (Array.isArray(query)) {
        return combineLatest(query.map((/**
         * @param {?} oneQuery
         * @return {?}
         */
        function (oneQuery) { return syncQuery.bind(_this, oneQuery); })));
    }
    if (!isQuery(query)) {
        throw new Error('Query should be either a path, a Query object or an array of Queries');
    }
    ////////////////
    // SUBQUERIES //
    ////////////////
    /**
     * Listen on Child actions
     * @type {?}
     */
    var fromChildAction = (/**
     * @param {?} actions
     * @param {?} child
     * @return {?}
     */
    function (actions, child) {
        var e_1, _a;
        /** @type {?} */
        var idKey = 'id';
        // TODO: Improve how to
        var parentId = child.parentId, key = child.key;
        var _loop_1 = function (action) {
            /** @type {?} */
            var id = action.payload.doc.id;
            /** @type {?} */
            var data = action.payload.doc.data();
            switch (action.type) {
                case 'added': {
                    _this['store'].update((/** @type {?} */ (parentId)), (/**
                     * @param {?} entity
                     * @return {?}
                     */
                    function (entity) {
                        var _a;
                        return (/** @type {?} */ ((_a = {},
                            _a[key] = arrayUpsert((/** @type {?} */ (entity[key])), id, data, idKey),
                            _a)));
                    }));
                    break;
                }
                case 'removed': {
                    _this['store'].update((/** @type {?} */ (parentId)), (/**
                     * @param {?} entity
                     * @return {?}
                     */
                    function (entity) {
                        var _a;
                        return (/** @type {?} */ ((_a = {},
                            _a[key] = arrayRemove((/** @type {?} */ (entity[key])), id, idKey),
                            _a)));
                    }));
                    break;
                }
                case 'modified': {
                    _this['store'].update((/** @type {?} */ (parentId)), (/**
                     * @param {?} entity
                     * @return {?}
                     */
                    function (entity) {
                        var _a;
                        return (/** @type {?} */ ((_a = {},
                            _a[key] = arrayUpdate((/** @type {?} */ (entity[key])), id, data, idKey),
                            _a)));
                    }));
                }
            }
        };
        try {
            for (var actions_1 = __values(actions), actions_1_1 = actions_1.next(); !actions_1_1.done; actions_1_1 = actions_1.next()) {
                var action = actions_1_1.value;
                _loop_1(action);
            }
        }
        catch (e_1_1) { e_1 = { error: e_1_1 }; }
        finally {
            try {
                if (actions_1_1 && !actions_1_1.done && (_a = actions_1.return)) _a.call(actions_1);
            }
            finally { if (e_1) throw e_1.error; }
        }
    });
    /**
     * Stay in sync a subquery of the collection
     * \@param subQuery The sub query based on the entity
     * \@param child An object that describe the parent
     * @type {?}
     */
    var syncSubQuery = (/**
     * @template K
     * @param {?} subQuery
     * @param {?} child
     * @return {?}
     */
    function (subQuery, child) {
        var _a;
        var parentId = child.parentId, key = child.key;
        // If it's a static value
        if (!isQuery(subQuery)) {
            /** @type {?} */
            var update = _this['store'].update((/** @type {?} */ (parentId)), (/** @type {?} */ (_a = {}, _a[key] = subQuery, _a)));
            return of(update);
        }
        if (Array.isArray(subQuery)) {
            /** @type {?} */
            var syncQueries = subQuery.map((/**
             * @param {?} oneQuery
             * @return {?}
             */
            function (oneQuery) {
                if (isQuery(subQuery)) {
                    /** @type {?} */
                    var id_1 = getIdAndPath({ path: subQuery.path });
                    return _this['db'].doc(subQuery.path).valueChanges().pipe(tap((/**
                     * @param {?} childDoc
                     * @return {?}
                     */
                    function (childDoc) {
                        _this['store'].update((/** @type {?} */ (parentId)), (/**
                         * @param {?} entity
                         * @return {?}
                         */
                        function (entity) {
                            var _a;
                            return (/** @type {?} */ ((_a = {},
                                _a[key] = arrayAdd((/** @type {?} */ (entity[key])), id_1, childDoc),
                                _a)));
                        }));
                    })));
                }
                return syncSubQuery(oneQuery, child);
            }));
            return combineLatest(syncQueries);
        }
        if (typeof subQuery !== 'object') {
            throw new Error('Query should be either a path, a Query object or an array of Queries');
        }
        // Sync subquery
        if (isDocPath(subQuery.path)) {
            return _this['db'].doc(subQuery.path).valueChanges().pipe(tap((/**
             * @param {?} children
             * @return {?}
             */
            function (children) {
                var _a;
                return _this['store'].update((/** @type {?} */ (parentId)), (/** @type {?} */ (_a = {}, _a[key] = children, _a)));
            })));
        }
        else {
            return (/** @type {?} */ (_this['db'].collection(subQuery.path, subQuery.queryFn)
                .stateChanges()
                .pipe(withTransaction((/**
             * @param {?} actions
             * @return {?}
             */
            function (actions) { return fromChildAction(actions, child); })))));
        }
    });
    /**
     * Get in sync with all the subqueries
     * \@param subQueries A map of all the subqueries
     * \@param parent The parent entity to attach the children to
     * @type {?}
     */
    var syncAllSubQueries = (/**
     * @param {?} subQueries
     * @param {?} parent
     * @return {?}
     */
    function (subQueries, parent) {
        /** @type {?} */
        var obs = Object.keys(subQueries)
            .filter((/**
         * @param {?} key
         * @return {?}
         */
        function (key) { return key !== 'path' && key !== 'queryFn'; }))
            .map((/**
         * @param {?} key
         * @return {?}
         */
        function (key) {
            /** @type {?} */
            var queryLike = getSubQuery(subQueries[key], parent);
            /** @type {?} */
            var child = { key: key, parentId: parent[_this.idKey] };
            return syncSubQuery(queryLike, child);
        }));
        return combineLatest(obs);
    });
    ////////////////
    // MAIN QUERY //
    ////////////////
    /**
     * Listen on action with child queries
     * @type {?}
     */
    var fromActionWithChild = (/**
     * @param {?} actions
     * @param {?} mainQuery
     * @param {?} subscriptions
     * @return {?}
     */
    function (actions, mainQuery, subscriptions) {
        var e_2, _a, _b;
        try {
            for (var actions_2 = __values(actions), actions_2_1 = actions_2.next(); !actions_2_1.done; actions_2_1 = actions_2.next()) {
                var action = actions_2_1.value;
                /** @type {?} */
                var id = action.payload.doc.id;
                /** @type {?} */
                var data = action.payload.doc.data();
                switch (action.type) {
                    case 'added': {
                        /** @type {?} */
                        var entity = __assign((_b = {}, _b[_this.idKey] = id, _b), data);
                        _this['store'].upsert(id, entity);
                        subscriptions[id] = syncAllSubQueries((/** @type {?} */ (mainQuery)), entity).subscribe();
                        break;
                    }
                    case 'removed': {
                        _this['store'].remove(id);
                        subscriptions[id].unsubscribe();
                        delete subscriptions[id];
                        break;
                    }
                    case 'modified': {
                        _this['store'].update(id, data);
                    }
                }
            }
        }
        catch (e_2_1) { e_2 = { error: e_2_1 }; }
        finally {
            try {
                if (actions_2_1 && !actions_2_1.done && (_a = actions_2.return)) _a.call(actions_2);
            }
            finally { if (e_2) throw e_2.error; }
        }
    });
    var path = query.path, queryFn = query.queryFn;
    if (isDocPath(path)) {
        var id_2 = this['getIdAndPath']({ path: path }).id;
        /** @type {?} */
        var subscription_1;
        return this['db'].doc(path).valueChanges().pipe(tap((/**
         * @param {?} entity
         * @return {?}
         */
        function (entity) {
            _this['store'].upsert(id_2, (/** @type {?} */ (__assign({ id: id_2 }, entity))));
            if (!subscription_1) { // Subscribe only the first time
                subscription_1 = syncAllSubQueries((/** @type {?} */ (query)), entity).subscribe();
            }
        })), 
        // Stop subscription
        finalize((/**
         * @return {?}
         */
        function () { return subscription_1.unsubscribe(); })));
    }
    else {
        /** @type {?} */
        var subscriptions_1 = {};
        return (/** @type {?} */ (this['db'].collection(path, queryFn)
            .stateChanges()
            .pipe(withTransaction((/**
         * @param {?} actions
         * @return {?}
         */
        function (actions) { return fromActionWithChild(actions, query, subscriptions_1); })), 
        // Stop all subscriptions
        finalize((/**
         * @return {?}
         */
        function () { return Object.keys(subscriptions_1).forEach((/**
         * @param {?} id
         * @return {?}
         */
        function (id) {
            subscriptions_1[id].unsubscribe();
            delete subscriptions_1[id];
        })); })))));
    }
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/**
 * Get the params from a path
 * @param {?} path
 * @return {?}
 */
function getPathParams(path) {
    return path.split('/')
        .filter((/**
     * @param {?} segment
     * @return {?}
     */
    function (segment) { return segment.charAt(0) === ':'; }))
        .map((/**
     * @param {?} segment
     * @return {?}
     */
    function (segment) { return segment.substr(1); }));
}
/**
 * Transform a path based on the params
 * \@example pathWithParams('movies/:movieId/stakeholder/:shId', { movieId, shId })
 * @param {?} path The path with params starting with "/:"
 * @param {?} params A map of id params
 * @return {?}
 */
function pathWithParams(path, params) {
    return path.split('/').map((/**
     * @param {?} segment
     * @return {?}
     */
    function (segment) {
        if (segment.charAt(0) === ':') {
            /** @type {?} */
            var key = segment.substr(1);
            if (!params[key]) {
                throw new Error("Required parameter " + key + " from " + path + " doesn't exist in params " + JSON.stringify(params));
            }
            return params[key];
        }
        else {
            return segment;
        }
    })).join('/');
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/**
 * Redirect the
 * @template E
 * @param {?} redirectTo Route path to redirecto if collection is empty
 * @return {?}
 */
function redirectIfEmpty(redirectTo) {
    return map((/**
     * @param {?} actions
     * @return {?}
     */
    function (actions) { return actions.length === 0 ? redirectTo : true; }));
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/**
 * @template S
 */
var  /**
 * @template S
 */
SubcollectionService = /** @class */ (function (_super) {
    __extends(SubcollectionService, _super);
    function SubcollectionService(store) {
        var _this = _super.call(this, store) || this;
        _this.store = store;
        try {
            _this.routerQuery = inject(RouterQuery);
        }
        catch (err) {
            throw new Error('SubCollectionService requires a "AkitaNgRouterStoreModule" from Akita.');
        }
        return _this;
    }
    Object.defineProperty(SubcollectionService.prototype, "path", {
        get: /**
         * @return {?}
         */
        function () {
            var _this = this;
            /** @type {?} */
            var pathParams = getPathParams(this.constructor['path']);
            return this.routerQuery.selectParams().pipe(
            // Don't trigger change if params needed (and only them) haven't changed
            distinctUntilChanged((/**
             * @param {?} old
             * @param {?} next
             * @return {?}
             */
            function (old, next) {
                return !pathParams.find((/**
                 * @param {?} param
                 * @return {?}
                 */
                function (param) { return old[param] !== next[param]; }));
            })), 
            // Need to filter because changes in params comes before canDeactivate
            filter((/**
             * @param {?} params
             * @return {?}
             */
            function (params) { return pathParams.every((/**
             * @param {?} param
             * @return {?}
             */
            function (param) { return !!params[param]; })); })), map((/**
             * @param {?} params
             * @return {?}
             */
            function (params) { return pathWithParams(_this.constructor['path'], params); })), tap((/**
             * @param {?} path
             * @return {?}
             */
            function (path) {
                if (path !== _this.collectionPath) {
                    _this.store.reset();
                    _this.collectionPath = path;
                }
            })), shareReplay(1));
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(SubcollectionService.prototype, "currentPath", {
        get: /**
         * @return {?}
         */
        function () {
            /** @type {?} */
            var params = this.routerQuery.getParams();
            return pathWithParams(this.constructor['path'], params);
        },
        enumerable: true,
        configurable: true
    });
    return SubcollectionService;
}(CollectionService));

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/**
 * @abstract
 * @template S
 */
var  /**
 * @abstract
 * @template S
 */
CollectionGroupService = /** @class */ (function () {
    function CollectionGroupService(store) {
        this.store = store;
        try {
            this.db = inject(AngularFirestore);
        }
        catch (err) {
            throw new Error('CollectionGroupService requires AngularFirestore.');
        }
    }
    Object.defineProperty(CollectionGroupService.prototype, "idKey", {
        get: /**
         * @return {?}
         */
        function () {
            return this.store.idKey;
        },
        enumerable: true,
        configurable: true
    });
    /** Sync the collection group with the store */
    /**
     * Sync the collection group with the store
     * @param {?=} queryGroupFn
     * @return {?}
     */
    CollectionGroupService.prototype.syncCollection = /**
     * Sync the collection group with the store
     * @param {?=} queryGroupFn
     * @return {?}
     */
    function (queryGroupFn) {
        return this.db.collectionGroup(this.collectionId, queryGroupFn).stateChanges()
            .pipe(withTransaction(syncFromAction.bind(this)));
    };
    /** Return a snapshot of the collection group */
    /**
     * Return a snapshot of the collection group
     * @param {?=} queryGroupFn
     * @return {?}
     */
    CollectionGroupService.prototype.getValue = /**
     * Return a snapshot of the collection group
     * @param {?=} queryGroupFn
     * @return {?}
     */
    function (queryGroupFn) {
        return __awaiter(this, void 0, void 0, function () {
            var snapshot;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0: return [4 /*yield*/, this.db.collectionGroup(this.collectionId, queryGroupFn).get().toPromise()];
                    case 1:
                        snapshot = _a.sent();
                        return [2 /*return*/, snapshot.docs.map((/**
                             * @param {?} doc
                             * @return {?}
                             */
                            function (doc) { return (/** @type {?} */ (doc.data())); }))];
                }
            });
        });
    };
    return CollectionGroupService;
}());

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */

export { CollectionConfig, CollectionGroupService, CollectionGuard, CollectionGuardConfig, CollectionService, SubcollectionService, collection, doc, getPathParams, isDocPath, pathWithParams, redirectIfEmpty, shoudCancel, syncFromAction, syncQuery, waitForCancel };
//# sourceMappingURL=akita-ng-fire.js.map
